<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Table;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Helper\Tfa as TfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\BackupcodesModel;
use Akeeba\Component\LoginGuard\Administrator\Service\Encrypt;
use Exception;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Event\DispatcherInterface;
use RuntimeException;

/**
 * Table for the Two-Step Verification second factor records
 *
 * @property int    $id          Record ID.
 * @property int    $user_id     User ID
 * @property string $title       Record title.
 * @property string $method      TFA Method (corresponds to one of the plugins).
 * @property int    $default     Is this the default Method?
 * @property array  $options     Configuration options for the TFA Method.
 * @property string $created_on  Date and time the record was created.
 * @property string $last_used   Date and time the record was las used successfully.
 *
 * @since    1.0.0
 */
class TfaTable extends AbstractTable
{
	/**
	 * Delete flags per ID, set up onBeforeDelete and used onAfterDelete
	 *
	 * @var   array
	 * @since 3.0.0
	 */
	private $deleteFlags = [];

	/**
	 * Internal flag used to create backup codes when I'm creating the very first TFA record
	 *
	 * @var   bool
	 * @since 3.0.0
	 */
	private $mustCreateBackupCodes = false;

	/**
	 * Encryption service
	 *
	 * @var   Encrypt
	 * @since 6.0.0
	 */
	private $encryptService;

	/**
	 * Table constructor
	 *
	 * @param   DatabaseDriver             $db           Database driver object
	 * @param   DispatcherInterface|null   $dispatcher   Events dispatcher object
	 */
	public function __construct(DatabaseDriver $db, DispatcherInterface $dispatcher = null)
	{
		parent::__construct('#__loginguard_tfa', 'id', $db, $dispatcher);

		$this->encryptService = new Encrypt();
	}

	/**
	 * Runs before storing a record to the database.
	 *
	 * - Encrypt the options
	 * - Make sure we have backup codes
	 * - Make sure there is a default option set
	 *
	 * @param   bool   $updateNulls
	 *
	 * @throws  Exception
	 * @since   3.0.0
	 */
	public function onBeforeStore(&$updateNulls)
	{
		$this->options = $this->encryptService->encrypt(json_encode($this->options ?: []));

		$this->mustCreateBackupCodes = 0;

		// We only care about new records, i.e. those with an ID that's NULL or 0
		if (!empty($this->id))
		{
			return;
		}

		// This does not apply to backup codes (otherwise we'd enter an infinite loop)
		if ($this->method == 'backupcodes')
		{
			return;
		}

		$records = TfaHelper::getUserTfaRecords($this->user_id);

		if ($this->id)
		{
			// Existing record. Remove it from the list of records.
			$records = array_filter($records, function ($rec) {
				return $rec->id != $this->id;
			});
		}
		else
		{
			// New record. Update the created_on column
			$this->created_on = Date::getInstance()->toSql();
		}

		// Do I need to mark this record as the default?
		if ($this->default == 0)
		{
			$hasDefaultRecord = array_reduce($records, function ($carry, $record) {
				return $carry || ($record->default == 1);
			}, false);

			if (!$hasDefaultRecord)
			{
				$this->default = 1;
			}
		}

		// Get the number of records this user_id has, except the current record.
		$numOldRecords = count($records);

		if ($numOldRecords == 0)
		{
			$this->mustCreateBackupCodes = 1;
			$this->default               = 1;
		}
	}

	/**
	 * Runs after saving the record to the database
	 *
	 * - Convert the encrypted options back to a decrypted array
	 * - Apply the default record preference
	 * - Generate backup codes if necessary
	 *
	 * @param   bool   $result
	 * @param   bool   $updateNulls
	 *
	 * @since   3.0.0
	 */
	public function onAfterStore(&$result, $updateNulls)
	{
		$this->decryptOptions();

		$this->switchDefaultRecord();

		$this->conditionallyRegenerateBackupCodes();

		$this->mustCreateBackupCodes = 0;
	}

	/**
	 * Decrypt the possibly encrypted options
	 *
	 * @since   3.0.0
	 */
	private function decryptOptions(): void
	{
		// Try with modern decryption
		$decrypted = @json_decode($this->encryptService->decrypt($this->options ?? ''), true);

		if (is_string($decrypted))
		{
			$decrypted = @json_decode($decrypted, true);
		}

		// FAll back to legacy decryption
		if (!is_array($decrypted))
		{
			$decrypted = @json_decode($this->encryptService->decrypt($this->options ?? '', true), true);

			if (is_string($decrypted))
			{
				$decrypted = @json_decode($decrypted, true);
			}
		}

		$this->options = $decrypted ?: [];
	}

	/**
	 * If this record is set to be the default, unset the default flag from the other records for the same user.
	 *
	 * @since  3.0.0
	 */
	private function switchDefaultRecord(): void
	{
		if (!$this->default)
		{
			return;
		}

		/**
		 * This record is marked as default, therefore we need to unset the default flag from all other records for this
		 * user.
		 */
		$db    = $this->getDbo();
		$zero  = 0;
		$query = $db->getQuery(true)
			->update($db->quoteName('#__loginguard_tfa'))
			->set($db->quoteName('default') . ' = :zero')
			->where($db->quoteName('user_id') . ' = :user_id')
			->where($db->quoteName('id') . ' != :id')
			->bind(':zero', $zero, ParameterType::INTEGER)
			->bind(':user_id', $this->user_id, ParameterType::INTEGER)
			->bind(':id', $this->id, ParameterType::INTEGER);
		$db->setQuery($query)->execute();
	}

	/**
	 * Regenerate backup code is the flag is set.
	 *
	 * @throws Exception
	 * @since  3.0.0
	 */
	private function conditionallyRegenerateBackupCodes(): void
	{
		if (!$this->mustCreateBackupCodes)
		{
			return;
		}

		/** @var MVCFactoryInterface $factory */
		$factory = Factory::getApplication()->bootComponent('com_loginguard')->getMVCFactory();

		/** @var BackupcodesModel $backupCodes */
		$backupCodes = $factory->createModel('Backupcodes', 'Administrator');
		$user        = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($this->user_id);
		$backupCodes->regenerateBackupCodes($user);
	}

	/**
	 * Runs after loading a new record.
	 *
	 * - Decrypt the possibly encrypted options
	 *
	 * @param   bool    $result   Did loading the record succeed
	 * @param   mixed   $keys     Key(s) used to load the record
	 * @param   bool    $reset    Was the table object reset on loading the new record?
	 *
	 * @since   3.0.0
	 */
	public function onAfterLoad(&$result, $keys, $reset)
	{
		// Only tun on successful table load
		if (!$result)
		{
			return;
		}

		$this->decryptOptions();
	}

	/**
	 * Runs before deleting a record
	 *
	 * - Collects information used to determine whether to change the default record and/or remove the backup codes
	 *
	 * @param   mixed   $pk
	 *
	 *
	 * @throws Exception
	 * @since  3.0.0
	 */
	public function onBeforeDelete(&$pk)
	{
		$record = $this;

		if ($pk != $this->id)
		{
			$record = clone $this;
			$record->reset();
			$result = $record->load($pk);

			if (!$result)
			{
				// If the record does not exist I will stomp my feet and deny your request
				throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
			}
		}

		$user = Factory::getApplication()->getIdentity()
			?? Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		// The user must be a registered user, not a guest
		if ($user->guest)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// You can only delete your own records, unless you're a super user or have delete privileges on this component
		if (($record->user_id != $user->id) && !TfaHelper::canEditUser($user))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Save flags used onAfterDelete
		$this->deleteFlags[$record->id] = [
			'default'    => $record->default,
			'numRecords' => $this->getNumRecords($record->user_id),
			'user_id'    => $record->user_id,
			'method'     => $record->method,
		];
	}

	/**
	 * Get the number of TFA records for a give user ID
	 *
	 * @param   int   $user_id   The user ID to check
	 *
	 * @return  int
	 *
	 * @since   6.0.0
	 */
	private function getNumRecords(int $user_id): int
	{
		$db            = $this->getDbo();
		$query         = $db->getQuery(true)
			->select('COUNT(*)')
			->from($db->quoteName('#__loginguard_tfa'))
			->where($db->quoteName('user_id') . ' = :user_id')
			->bind(':user_id', $user_id, ParameterType::INTEGER);
		$numOldRecords = $db->setQuery($query)->loadResult();

		return (int) $numOldRecords;
	}

	/**
	 * Runs after deleting a record
	 *
	 * @param   bool    $result
	 * @param   mixed   $pk
	 *
	 * @since   3.0.0
	 */
	public function onAfterDelete(&$result, $pk)
	{
		if (!$result)
		{
			return;
		}

		if (is_array($pk))
		{
			$pk = array_shift($pk);
		}

		if (!isset($this->deleteFlags[$pk]))
		{
			return;
		}

		if (($this->deleteFlags[$pk]['numRecords'] <= 2) && ($this->deleteFlags[$pk]['method'] != 'backupcodes'))
		{
			/**
			 * This was the second to last TFA record in the database (the last one is the `backupcodes`). Therefore, we
			 * need to delete the remaining entry and go away. We don't trigger this if the Method we are deleting was
			 * the `backupcodes` because we might just be regenerating the backup codes.
			 */
			$db    = $this->getDbo();
			$query = $db->getQuery(true)
				->delete($db->quoteName('#__loginguard_tfa'))
				->where($db->quoteName('user_id') . ' = :user_id')
				->bind(':user_id', $this->deleteFlags[$pk]['user_id'], ParameterType::INTEGER);
			$db->setQuery($query)->execute();

			unset($this->deleteFlags[$pk]);

			return;
		}

		// This was the default record. Promote the next available record to default.
		if ($this->deleteFlags[$pk]['default'])
		{
			$db    = $this->getDbo();
			$query = $db->getQuery(true)
				->select($db->quoteName('id'))
				->from($db->quoteName('#__loginguard_tfa'))
				->where($db->quoteName('user_id') . ' = :user_id')
				->bind(':user_id', $this->deleteFlags[$pk]['user_id'], ParameterType::INTEGER);
			$ids   = $db->setQuery($query)->loadColumn();

			if (empty($ids))
			{
				return;
			}

			$id    = array_shift($ids);
			$one   = 1;
			$query = $db->getQuery(true)
				->update($db->quoteName('#__loginguard_tfa'))
				->set($db->qn('default') . ' = :one')
				->where($db->quoteName('id') . ' = :id')
				->bind(':one', $one, ParameterType::INTEGER)
				->bind(':id', $id, ParameterType::INTEGER);
			$db->setQuery($query)->execute();
		}
	}
}