<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Loginguard\Fixed\Extension;

// Prevent direct access
use Akeeba\Component\LoginGuard\Administrator\DataShape\CaptiveRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\GetInfo;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowCaptive;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\User\User;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use Joomla\Input\Input;
use RuntimeException;
use stdClass;

defined('_JEXEC') || die;

/**
 * Akeeba LoginGuard Plugin for Two Step Verification Method "Fixed"
 *
 * Requires a static string (password), different for each user. It effectively works as a second password. This is NOT
 * to be used on production sites. It serves as a demonstration plugin and as a template for developers to create their
 * own custom Two Step Verification plugins.
 */
class Fixed extends CMSPlugin implements SubscriberInterface
{
	/**
	 * Forbid registration of legacy (Joomla 3) event listeners.
	 *
	 * @var    boolean
	 * @since  6.0.0
	 *
	 * @deprecated
	 */
	protected $allowLegacyListeners = false;

	/**
	 * Autoload this plugin's language files
	 *
	 * @var    boolean
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The TFA Method name handled by this plugin
	 *
	 * @var   string
	 */
	private $tfaMethodName = 'fixed';

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onLoginGuardTfaGetMethod' => 'onLoginGuardTfaGetMethod',
			'onLoginGuardTfaCaptive'   => 'onLoginGuardTfaCaptive',
			'onLoginGuardTfaGetSetup'  => 'onLoginGuardTfaGetSetup',
			'onLoginGuardTfaSaveSetup' => 'onLoginGuardTfaSaveSetup',
			'onLoginGuardTfaValidate'  => 'onLoginGuardTfaValidate',
		];
	}

	/**
	 * Gets the identity of this TFA Method
	 *
	 * @param   GetInfo  $event  The event we are handling
	 *
	 * @return  void
	 */
	public function onLoginGuardTfaGetMethod(GetInfo $event): void
	{
		$event->addResult(
			new MethodDescriptor([
				'name'      => $this->tfaMethodName,
				'display'   => Text::_('PLG_LOGINGUARD_FIXED_LBL_DISPLAYEDAS'),
				'shortinfo' => Text::_('PLG_LOGINGUARD_FIXED_LBL_SHORTINFO'),
				'image'     => 'media/plg_loginguard_fixed/images/fixed.svg',
				'help_url'  => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Fixed-Code'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the Captive TFA page. This is the page which appears
	 * right after you log in and asks you to validate your login with TFA.
	 *
	 * @param   ShowCaptive  $event  The event we are handling
	 *
	 * @return  void
	 */
	public function onLoginGuardTfaCaptive(ShowCaptive $event): void
	{
		/**
		 * @var   stdClass $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		$event->addResult(
			new CaptiveRenderOptions([
				// Custom HTML to display above the TFA form
				'pre_message'  => Text::_('PLG_LOGINGUARD_FIXED_LBL_PREMESSAGE'),
				// How to render the TFA code field. "input" (HTML input element) or "custom" (custom HTML)
				'field_type'   => 'input',
				// The type attribute for the HTML input box. Typically "text" or "password". Use any HTML5 input type.
				'input_type'   => 'password',
				// Placeholder text for the HTML input box. Leave empty if you don't need it.
				'placeholder'  => Text::_('PLG_LOGINGUARD_FIXED_LBL_PLACEHOLDER'),
				// Label to show above the HTML input box. Leave empty if you don't need it.
				'label'        => Text::_('PLG_LOGINGUARD_FIXED_LBL_LABEL'),
				// Custom HTML. Only used when field_type = custom.
				'html'         => '',
				// Custom HTML to display below the TFA form
				'post_message' => Text::_('PLG_LOGINGUARD_FIXED_LBL_POSTMESSAGE'),
				// URL for help content
				'help_url'     => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Fixed-Code'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the TFA setup page. This is the page which allows the
	 * user to add or modify a TFA Method for their user account. If the record does not correspond to your plugin
	 * return an empty array.
	 *
	 * @param   ShowSetup  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardTfaGetSetup(ShowSetup $event): void
	{
		/** @var TfaTable $record The #__loginguard_tfa record currently selected by the user. */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$helpURL = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Fixed-Code');

		/**
		 * Return the parameters used to render the GUI.
		 *
		 * Some TFA Methods need to display a different interface before and after the setup. For example, when setting
		 * up Google Authenticator or a hardware OTP dongle you need the user to enter a TFA code to verify they are in
		 * possession of a correctly configured device. After the setup is complete you don't want them to see that
		 * field again. In the first state you could use the tabular_data to display the setup values, pre_message to
		 * display the QR code and field_type=input to let the user enter the TFA code. In the second state do the same
		 * BUT set field_type=custom, set html='' and show_submit=false to effectively hide the setup form from the
		 * user.
		 */
		$event->addResult(
			new SetupRenderOptions([
				'default_title' => Text::_('PLG_LOGINGUARD_FIXED_LBL_DEFAULTTITLE'),
				'pre_message'   => Text::_('PLG_LOGINGUARD_FIXED_LBL_SETUP_PREMESSAGE'),
				'field_type'    => 'input',
				'input_type'    => 'password',
				'input_value'   => $options->fixed_code,
				'placeholder'   => Text::_('PLG_LOGINGUARD_FIXED_LBL_PLACEHOLDER'),
				'label'         => Text::_('PLG_LOGINGUARD_FIXED_LBL_LABEL'),
				'post_message'  => Text::_('PLG_LOGINGUARD_FIXED_LBL_SETUP_POSTMESSAGE'),
				'help_url'      => $helpURL,
			])
		);
	}

	/**
	 * Parse the input from the TFA setup page and return the configuration information to be saved to the database. If
	 * the information is invalid throw a RuntimeException to signal the need to display the editor page again. The
	 * message of the exception will be displayed to the user. If the record does not correspond to your plugin return
	 * an empty array.
	 *
	 * @param   SaveSetup  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardTfaSaveSetup(SaveSetup $event): void
	{
		/**
		 * @var TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 * @var Input    $input  The user input you are going to take into account.
		 */
		$record = $event['record'];
		$input  = $event['input'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);

		// Merge with the submitted form data
		$code = $input->get('code', $options->fixed_code, 'raw');

		// Make sure the code is not empty
		if (empty($code))
		{
			throw new RuntimeException(Text::_('PLG_LOGINGUARD_FIXED_ERR_EMPTYCODE'));
		}

		// Return the configuration to be serialized
		$event->addResult([
			'fixed_code' => $code,
		]);
	}

	/**
	 * Validates the Two Factor Authentication code submitted by the user in the Captive Two Step Verification page. If
	 * the record does not correspond to your plugin return FALSE.
	 *
	 * @param   Validate  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardTfaValidate(Validate $event): void
	{
		/**
		 * @var   TfaTable    $record The TFA Method's record you're validating against
		 * @var   User        $user   The user record
		 * @var   string|null $code   The submitted code
		 */
		$record = $event['record'];
		$user   = $event['user'];
		$code   = $event['code'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			$event->addResult(false);

			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);

		// Double check the TFA Method is for the correct user
		if ($user->id != $record->user_id)
		{
			$event->addResult(false);

			return;
		}

		// Check the TFA code for validity
		$event->addResult(hash_equals($options->fixed_code, $code ?? ''));
	}

	/**
	 * Decodes the options from a #__loginguard_tfa record into an options object.
	 *
	 * @param   TfaTable  $record
	 *
	 * @return  object
	 */
	private function _decodeRecordOptions(TfaTable $record): object
	{
		$options = [
			'fixed_code' => '',
		];

		if (!empty($record->options))
		{
			$recordOptions = $record->options;

			$options = array_merge($options, $recordOptions);
		}

		return (object) $options;
	}
}
