<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Loginguard\Totp\Extension;

// Prevent direct access
defined('_JEXEC') || die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\CaptiveRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\GetInfo;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowCaptive;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Encrypt\Totp as TotpHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Input\Input;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use RuntimeException;

/**
 * Akeeba LoginGuard Plugin for Two Step Verification Method "Time-based One Time Password"
 *
 * Requires a 6-digit code generated by Google Authenticator or any compatible application. These codes change
 * automatically every 30 seconds.
 */
class Totp extends CMSPlugin implements SubscriberInterface
{
	/**
	 * The application we are running under.
	 *
	 * @var    CMSApplication
	 * @since  1.1.0
	 */
	protected $app;

	/**
	 * Auto-load the language files
	 *
	 * @var    bool
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The TFA Method name handled by this plugin
	 *
	 * @var   string
	 */
	private $tfaMethodName = 'totp';

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onLoginGuardTfaGetMethod' => 'onLoginGuardTfaGetMethod',
			'onLoginGuardTfaCaptive'   => 'onLoginGuardTfaCaptive',
			'onLoginGuardTfaGetSetup'  => 'onLoginGuardTfaGetSetup',
			'onLoginGuardTfaSaveSetup' => 'onLoginGuardTfaSaveSetup',
			'onLoginGuardTfaValidate'  => 'onLoginGuardTfaValidate',
		];
	}

	/**
	 * Gets the identity of this TFA Method
	 *
	 * @param   GetInfo  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onLoginGuardTfaGetMethod(GetInfo $event): void
	{
		$event->addResult(
			new MethodDescriptor([
				'name'      => $this->tfaMethodName,
				'display'   => Text::_('PLG_LOGINGUARD_TOTP_LBL_DISPLAYEDAS'),
				'shortinfo' => Text::_('PLG_LOGINGUARD_TOTP_LBL_SHORTINFO'),
				'image'     => 'media/plg_loginguard_totp/images/totp.svg',
				'help_url'  => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Authenticator-App'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the Captive TFA page. This is the page which appears
	 * right after you log in and asks you to validate your login with TFA.
	 *
	 * @param   ShowCaptive  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onLoginGuardTfaCaptive(ShowCaptive $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		$event->addResult(
			new CaptiveRenderOptions([
				// Custom HTML to display above the TFA form
				'pre_message'  => '',
				// How to render the TFA code field. "input" (HTML input element) or "custom" (custom HTML)
				'field_type'   => 'input',
				// The type attribute for the HTML input box. Typically "text" or "password". Use any HTML5 input type.
				'input_type'   => 'number',
				// Placeholder text for the HTML input box. Leave empty if you don't need it.
				'placeholder'  => '',
				// Label to show above the HTML input box. Leave empty if you don't need it.
				'label'        => Text::_('PLG_LOGINGUARD_TOTP_LBL_LABEL'),
				// Custom HTML. Only used when field_type = custom.
				'html'         => '',
				// Custom HTML to display below the TFA form
				'post_message' => '',
				// URL for help content
				'help_url'     => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Authenticator-App'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the TFA setup page. This is the page which allows the
	 * user to add or modify a TFA Method for their user account. If the record does not correspond to your plugin
	 * return an empty array.
	 *
	 * @param   ShowSetup  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onLoginGuardTfaGetSetup(ShowSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		$totp = new TotpHelper();

		// Load the options from the record (if any)
		$options      = $this->_decodeRecordOptions($record);
		$key          = $options['key'] ?? '';
		$session      = $this->app->getSession();
		$isConfigured = !empty($key);

		// If there's a key in the session use that instead.
		$sessionKey = $session->get('com_loginguard.totp.key', null);

		if (!empty($sessionKey))
		{
			$key = $sessionKey;
		}

		// If there's still no key in the options, generate one and save it in the session
		if (empty($key))
		{
			$key = $totp->generateSecret();
			$session->set('com_loginguard.totp.key', $key);
		}

		// Generate a QR code for the key
		$user     = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($record->user_id);
		$hostname = Uri::getInstance()->toString(['host']);
		$otpURL   = sprintf("otpauth://totp/%s@%s?secret=%s", $user->username, $hostname, $key);
		$helpURL  = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Authenticator-App');
		$document = $this->app->getDocument();
		$wam      = $document->getWebAssetManager();

		$document->addScriptOptions('plg_loginguard.totp.qr', $otpURL);

		$wam->getRegistry()->addExtensionRegistryFile('plg_loginguard_totp');
		$wam->useScript('plg_loginguard_totp.setup');

		$event->addResult(
			new SetupRenderOptions([
				'default_title' => Text::_('PLG_LOGINGUARD_TOTP_LBL_DISPLAYEDAS'),
				'pre_message'   => Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_INSTRUCTIONS'),
				'table_heading' => Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_HEADING'),
				'tabular_data'  => [
					''                                                  => '<h5>' . Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_SUBHEAD') . '</h5>',
					Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_KEY')  => $key,
					Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_QR')   => "<span id=\"loginGuardQRImage\" />",
					Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_LINK') => Text::sprintf('PLG_LOGINGUARD_TOTP_LBL_SETUP_TABLE_LINK_TEXT', $otpURL),
				],
				'hidden_data'   => [
					'key' => $key,
				],
				'field_type'    => $isConfigured ? 'custom' : 'input',
				'input_type'    => 'number',
				'input_value'   => '',
				'placeholder'   => Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_PLACEHOLDER'),
				'label'         => Text::_('PLG_LOGINGUARD_TOTP_LBL_SETUP_LABEL'),
				'help_url'      => $helpURL,
			])
		);
	}

	/**
	 * Parse the input from the TFA setup page and return the configuration information to be saved to the database. If
	 * the information is invalid throw a RuntimeException to signal the need to display the editor page again. The
	 * message of the exception will be displayed to the user. If the record does not correspond to your plugin return
	 * an empty array.
	 *
	 * @param   SaveSetup  $event
	 *
	 * @return  void The configuration data to save to the database
	 * @since   6.0.0
	 */
	public function onLoginGuardTfaSaveSetup(SaveSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 * @var   Input    $input  The user input you are going to take into account.
		 */
		$record = $event['record'];
		$input  = $event['input'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options    = $this->_decodeRecordOptions($record);
		$optionsKey = $options['key'] ?? '';
		$key        = $optionsKey;
		$session    = $this->app->getSession();

		// If there is no key in the options fetch one from the session
		if (empty($key))
		{
			$key = $session->get('com_loginguard.totp.key', null);
		}

		// If there is still no key in the options throw an error
		if (empty($key))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		/**
		 * If the code is empty but the key already existed in $options someone is simply changing the title / default
		 * Method status. We can allow this and stop checking anything else now.
		 */
		$code = $input->getInt('code');

		if (empty($code) && !empty($optionsKey))
		{
			$event->addResult($options);

			return;
		}

		// In any other case validate the submitted code
		$totp    = new TotpHelper();
		$isValid = $totp->checkCode($key, $code);

		if (!$isValid)
		{
			throw new RuntimeException(Text::_('PLG_LOGINGUARD_TOTP_ERR_INVALID_CODE'), 500);
		}

		// The code is valid. Unset the key from the session.
		$session->set('com_loginguard.totp.key', null);

		// Return the configuration to be serialized
		$event->addResult([
			'key' => $key,
		]);
	}

	/**
	 * Validates the Two Factor Authentication code submitted by the user in the Captive Two Step Verification page. If
	 * the record does not correspond to your plugin return FALSE.
	 *
	 * @param   Validate  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onLoginGuardTfaValidate(Validate $event): void
	{
		/**
		 * @var   TfaTable $record The TFA Method's record you're validatng against
		 * @var   User     $user   The user record
		 * @var   string   $code   The submitted code
		 */
		$record = $event['record'];
		$user   = $event['user'];
		$code   = $event['code'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			$event->addResult(false);

			return;
		}

		// Double check the TFA Method is for the correct user
		if ($user->id != $record->user_id)
		{
			$event->addResult(false);

			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$key     = $options['key'] ?? '';

		// If there is no key in the options throw an error
		if (empty($key))
		{
			$event->addResult(false);

			return;
		}

		// Check the TFA code for validity
		$event->addResult((new TotpHelper())->checkCode($key, $code));
	}

	/**
	 * Decodes the options from a #__loginguard_tfa record into an options object.
	 *
	 * @param   TfaTable  $record
	 *
	 * @return  array
	 * @since   1.0.0
	 */
	private function _decodeRecordOptions(TfaTable $record): array
	{
		$options = [
			'key' => '',
		];

		if (!empty($record->options))
		{
			$recordOptions = $record->options;

			$options = array_merge($options, $recordOptions);
		}

		return $options;
	}
}
